#!/usr/bin/env node

/*
 blackcatmq
 copyright (c) 2012 Yaroslav Gaponov <yaroslav.gaponov@gmail.com>
*/

var DEBUG = true;

var os = require('os');
var util = require('util');
var net = require('net');
var fs = require('fs');

var stomp = require('./lib/stomp.js');

var sender = function(socket, data) {
    if (data) {
        if (socket && socket.writable) {
            socket.write(data.toString());
        }
    }
}

var BlackCatMQ = function(config) {
    var self = this;
    
    if (self instanceof BlackCatMQ) {        
        self.port = config.port || 61613;
        self.host = config.host || '0.0.0.0';
        self.interval = config.interval || 50000;
        
        self.sockets = {};
        self.subscribes = {};
        
        self.messages = { frame: {}, queue: [] };
        self.ack_list = [];
        
        self.transactions = {};
        
        self.server = net.createServer({ allowHalfOpen: true }, function(socket) {            
            socket.setEncoding('utf8');
            socket.setKeepAlive(true);
            
            var remoteAddress = socket.address();
            util.log(util.format('server is connected to %s:%s', remoteAddress.address, remoteAddress.port));
            
            var data = '';
            socket.on('data', function(chunk) {
                
                fs.appendFileSync('./dump/' + os.platform() + '.dat', chunk, encoding='utf8');
                
                data += chunk.toString();
                
                var frames = data.split(stomp.DELIMETER);
                                
                if (frames.length > 1) {
                    data = frames.pop();                
                    frames.forEach(function(_frame) {
                        var frame = stomp.Frame(_frame);
                        try {
                            
                            if (DEBUG) util.log(frame.command);
                            
                            if (self[frame.command.toLowerCase()] && typeof self[frame.command.toLowerCase()] === 'function') {
                                sender(socket, self[frame.command.toLowerCase()].call(self, socket, frame));
                            } else {
                                sender(socket, stomp.ServerFrame.ERROR('invalid parameters','command ' + frame.command + ' is not supported'));
                            }
                        } catch (ex) {
                            sender(socket, stomp.ServerFrame.ERROR(ex, ex.stack));
                        }
                    });
                }   
                
            });
            
            socket.on('end', function() {
                util.log(util.format('server is disconnected from %s:%s', remoteAddress.address, remoteAddress.port));                
            });
        });        
    } else {
        return new BlackCatMQ(config);
    }
}

/*
 start server
*/
BlackCatMQ.prototype.start = function(callback) {
    var self = this;
    
    self.server.listen(self.port, self.host, function() {
        var addr = self.server.address();
        util.log(util.format("server is started on %s:%s ...", addr.address, addr.port));
        
        self.timerID = setInterval(function() { self.timer(); }, self.interval);
        
        if (callback && typeof callback === 'function') {
            return callback();
        }
    });    
}

/*
 stop server
*/
BlackCatMQ.prototype.stop = function(callback) {
    var self = this;
    
    self.server.close(function() {
        util.log('server is stopped');
        
        clearInterval(self.timerID);
        
        if (callback && typeof callback === 'function') {
            return callback();
        }    
    });    
}

/*
 STOMP command  -> connect
*/
BlackCatMQ.prototype.connect = function(socket, frame) {
    var self = this;
    
    var session = 'id' + Math.floor(Math.random()*999999999999999999999);
    self.sockets[session] = socket;
    return stomp.ServerFrame.CONNECTED(session);
}

/*
 STOMP command -> subscribe
*/
BlackCatMQ.prototype.subscribe = function(socket, frame) {
    var self = this;
    
    if (! frame.header) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no header section');
    }
    
    var destination = frame.header['destination'];
    if (! destination) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no destination argument');
    }
    
    var session = frame.header['session'];
    if (! session) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no session argument');
    }
    
    
    if (frame.header['ack'] && frame.header['ack'] === 'client') {
        self.ack_list.push(session);
    }
    
    if (self.subscribes[destination]) {
        self.subscribes[destination].push(session);    
    } else {
        self.subscribes[destination] = [session];   
    }
}

/*
 STOMP command -> unsubsctibe
*/
BlackCatMQ.prototype.unsubscribe = function(socket, frame) {
    var self = this;

    if (! frame.header) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no header section');
    }
    
    var destination = frame.header['destination'];
    if (! destination) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no destination argument');
    }
    
    var session = frame.header['session'];
    if (! session) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no session argument');
    }
    
    if (self.subscribes[destination]) {
        var pos = self.subscribes[destination].indexOf(session);
        if (pos >= 0) {
            self.subscribes[destination].splice(pos, 1);   
        }        
    }
}

/*
 STOMP command -> send
*/
BlackCatMQ.prototype.send = function(socket, frame) {
    var self = this;
    
    if (! frame.header) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no header section');
    }
    
    var destination = frame.header['destination'];
    if (! destination) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no destination argument');
    }
    
    var transaction = frame.header['transaction'];
    if (transaction) {
        self.transactions[transaction].push(frame);
    }
    
    if (self.subscribes[destination]) {
                
        var messageID = 'id' + Math.floor(Math.random()*999999999999999999999);
                
        if (destination.indexOf('/queue/') === 0) {
            var session =  self.subscribes[destination].pop();
            self.subscribes[destination].unshift(session);
                        
            if (self.ack_list.indexOf(session) >= 0) {
                self.messages.frame[messageID] = frame;
                self.messages.queue.push(messageID);
            }            
            sender(self.sockets[session], stomp.ServerFrame.MESSAGE(destination, messageID, frame.body));            
        } else {
            self.subscribes[destination].forEach(function(session) {
                sender(self.sockets[session], stomp.ServerFrame.MESSAGE(destination, messageID, frame.body));                
            });    
        }
    }
}

/*
 STOMP command -> ack
*/
BlackCatMQ.prototype.ack = function(socket, frame) {
    var self = this;
    
    if (! frame.header) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no header section');
    }
    
    var messageID = frame.header['message-id'];
    if (! messageID) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no message-id argument');
    }
        
    delete self.messages.frame[messageID];
    var pos = self.messages.queue.indexOf(messageID);    
    if (pos >= 0) {
        self.messages.queue.splice(pos, 1);
    }
    
    return stomp.ServerFrame.RECEIPT(messageID);
}


/*
 STOMP command -> disconnect
*/
BlackCatMQ.prototype.disconnect = function(socket, frame) {
    var self = this;
    
    for(var session in self.sockets) {
        if (self.sockets[session] === socket) {
            delete self.sockets[session];
            
            for (var destination in self.subscribes) {
                var pos = self.subscribes[destination].indexOf(session);
                if (pos >= 0) {
                    self.subscribes[destination].splice(pos, 1);   
                }
            }
            
            var pos = self.ack_list.indexOf(session);
            if (pos >= 0) {
                self.ack_list.splice(pos, 1);
            }
            break;
        }        
    }    
}

/*
 STOMP command -> begin
*/
BlackCatMQ.prototype.begin = function(socket, frame) {
    var self = this;

    if (! frame.header) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no header section');
    }
    
    var trasaction = frame.header['transaction'];
    if (! trasaction) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no transaction argument');
    }
    
    self.transactions[transaction] = [];
}

/*
 STOMP command -> commit
*/
BlackCatMQ.prototype.commit = function(socket, frame) {
    var self = this;

    if (! frame.header) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no header section');
    }

    var trasaction = frame.header['transaction'];
    if (! trasaction) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no transaction argument');
    }
    
    delete self.transactions[transaction];
}

/*
 STOMP command -> abort
*/
BlackCatMQ.prototype.abort = function(socket, frame) {
    var self = this;

    if (! frame.header) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no header section');
    }
    
    var trasaction = frame.header['transaction'];
    if (! trasaction) {
        return stomp.ServerFrame.ERROR('invalid parameters','there is no transaction argument');
    }
    
    self.transactions[transaction].forEach(function(frame) {
        self.send(null, frame);
    });    
    delete self.transactions[transaction];
}


/*
 periodic task - return of lost messages 
*/
BlackCatMQ.prototype.timer = function() {
    var self = this;
    
    if (self.messages.queue.length > 0) {
        var messageID = self.messages.queue.shift();
        self.send(null, self.messages.frame[messageID]);
        delete self.messages.frame[messageID];
    }
}


/*
 initalize & run server
*/
fs.readFile(__dirname + '/blackcatmq.conf', 'utf8', function(err, data) {    
    if (err) throw err;
    
    var config = JSON.parse(data);

    var server = new BlackCatMQ(config);
    if (server) {
        server.start();
    }
    
    process.once('uncaughtException', function(err) {
        util.debug('error:' + err + err.stack);
        if (server) {
            server.stop();
        }
    });
    
    process.once('exit', function() {
        if (server) {
            server.stop();
        }        
    });
    
    process.once('SIGINT', function() {
        if (server) {
            server.stop();
        }
        console.log('Got SIGINT.  Press Control-c to exit.');
    });    
});


